package edu.berkeley.sbp;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;

/** an element which matches exactly one input token */
public abstract class Atom<T> extends Element implements Topology<T> {

    protected abstract Topology<T> top();
    public    abstract String       toString();
    public             StringBuffer toString(StringBuffer sb) { sb.append(this); return sb; }

    // Topology Thunks //////////////////////////////////////////////////////////////////////////////

    public Topology<T>       unwrap()                   { return top().unwrap(); }
    public Topology<T>       empty()                    { return top().empty(); }
    public boolean           contains(T v)              { return top().contains(v); }
    public Topology<T>       intersect(Topology<T> t)   { return top().intersect(t); }
    public Topology<T>       minus(Topology<T> t)       { return top().minus(t); }
    public Topology<T>       union(Topology<T> t)       { return top().union(t); }
    public Topology<T>       complement()               { return top().complement(); }
    public boolean           disjoint(Topology<T> t)    { return top().disjoint(t); }
    public boolean           containsAll(Topology<T> t) { return top().containsAll(t); }
    public int               hashCode()                 { return top().hashCode(); }
    public boolean           equals(Object o)           { return o != null && o instanceof Atom && ((Atom)o).top().equals(top()); }

    /** if all expressions matching <tt>e</tt> are exactly one token
     *  long, <b>attempt to</b> return an Atom representing that token
     *  (undecidable in general; only works in trivial cases)
     */
    public static Topology toAtom(Element e) {
        if (e instanceof Atom) return (Atom)e;
        if (e instanceof Sequence) return ((Sequence)e).toAtom();
        Topology ret = null;
        for(Sequence s : (Union)e)
            ret = ret==null ? toAtom(s) : ret.union(s.toAtom());
        return ret;
    }
}

